import { FrontOfficeTicketData } from "./frontOfficeTicketData";
import { FrontOfficeExecuteActionDataAction, FrontOfficeExecuteActionDataActionResponse } from "./frontOfficeExecuteActionData";

export class FrontOfficeClient {
    private readonly approvalProcessEndpoint = "approvalProcesses/executeAction";
    private readonly ticketsEndpoint = "tickets";
    private readonly maxResponseLength = 3000;

    constructor(private readonly connector: string) {
        this.connector = connector;
    }

    /**
     * Creates JSON object with mandatory parameters representing body that can be send as POST to /tickets endpoint.
     * @return JSON object
     */
    getSingleTicketData(
        stateId: string,
        holderType: string,
        holderValue: string,
        template: string,
        attachmentFromGlobalStorageEnabled: boolean = false,
        uploadAttachmentFromLocalDriveEnabled: boolean = false,
    ): FrontOfficeTicketData {
        return {
            type: "ticket",
            stateId: stateId,
            holder: {
                type: holderType,
                holder: holderValue,
            },
            documentData: {
                templateName: template,
            },
            addAttachmentFromGlobalStorageEnabled: attachmentFromGlobalStorageEnabled,
            uploadAttachmentFromLocalDriveEnabled: uploadAttachmentFromLocalDriveEnabled,
        };
    }

    /**
     * Creates Front Office ticket.
     * @param data JSON object that will be given as body
     * @returns Promise of ID of created ticket.
     */
    async createSingleTicket(data: FrontOfficeTicketData): Promise<string> {
        const response = await this.fetchFrontOffice(this.ticketsEndpoint, "POST", data);
        const json = await response.json();

        if (json.guid === undefined) {
            throw new Error("Could not parse ID of created ticket.");
        }

        return json.guid;
    }

    /**
     * Creates approval process action.
     * @actionId The approval-process action's ID as defined in the approval process.
     * @guid The ticket's GUID.
     * @message Custom message about the approval-process action.
     * @return JSON object
     */
    getApprovalProcessesExecuteActionData(actionId: string, guid: string, message: string): FrontOfficeExecuteActionDataAction {
        return {
            actionId: actionId,
            guid: guid,
            message: message,
            objectType: "Ticket",
        };
    }

    /**
     * Executes approval process actions.
     * @frontOfficeExecuteActionDataActions The approval process action that moves the ticket to the desired state.
     */
    async approvalProcessesExecuteActions(frontOfficeExecuteActionDataActions: FrontOfficeExecuteActionDataAction[]): Promise<void> {
        console.log(`Executing '${frontOfficeExecuteActionDataActions.length}' approval processes actions.`);
        for (const frontOfficeExecuteActionDataAction of frontOfficeExecuteActionDataActions) {
            const response = await this.fetchApprovalProcessesExecuteAction(frontOfficeExecuteActionDataAction);
            if (response.remaining == 0) {
                const waitingSeconds = this.getWaitingTime(response.resetTimeStamp);
                console.log(`Request limit reached. Waiting for '${waitingSeconds}' seconds`);
                await this.wait(waitingSeconds);
                console.log(`Continuing sending requests`);
            }
        }
        console.log(`Successfully executed '${frontOfficeExecuteActionDataActions.length}' approval processes actions.`);
    }

    private async fetchApprovalProcessesExecuteAction(
        data: FrontOfficeExecuteActionDataAction,
    ): Promise<FrontOfficeExecuteActionDataActionResponse> {
        const response = await this.fetchFrontOffice(this.approvalProcessEndpoint, "POST", data);

        const result: FrontOfficeExecuteActionDataActionResponse = {
            statusCode: response.status,
            resetTimeStamp: Number(response.headers.get("X-Ratelimit-Reset")!),
            rateLimit: Number(response.headers.get("X-Ratelimit-Limit")),
            remaining: Number(response.headers.get("X-Ratelimit-Remaining")),
        };

        if (!response.ok) {
            throw new Error(`Web endpoint error: HTTP Code ${response.status} - '${await response.text()}'.`);
        }

        await response.close();
        return result;
    }

    private async fetchFrontOffice(relativePath: string, method: string, data: any): Promise<any> {
        const url = this.connector + relativePath;

        const response = await fetch(url, {
            method: method,
            headers: {
                "Content-Type": "application/json",
            },
            body: JSON.stringify(data),
        });

        if (!response.ok) {
            const responseText = await response.text();
            const truncatedResponseText =
                responseText.length > this.maxResponseLength ? responseText.substring(0, this.maxResponseLength) + "..." : responseText;
            throw new Error(`Web endpoint error: HTTP Code ${response.status} - '${truncatedResponseText}'.`);
        }

        return response;
    }

    private getWaitingTime(epochSeconds: number): number {
        const now = new Date();
        const nowSeconds = Math.floor(now.getTime() / 1000);

        return epochSeconds - nowSeconds;
    }

    private async wait(s: number): Promise<void> {
        const ms = s * 1000;
        return new Promise((resolve) => setTimeout(resolve, ms));
    }
}
